import {
  h,
  ref,
  watch,
  toRef,
  Teleport,
  reactive,
  onMounted,
  onUnmounted,
  defineComponent,
} from "vue";

declare class HyperFunctionComponent {
  constructor(container: HTMLDivElement, props: HfcProps);
  static interface: {
    attrs?: Record<string, any>;
    events?: Record<string, any>;
    slots?: Record<string, any>;
  };

  connectedCallback(): void;
  attributeChangedCallback(
    type: "attr" | "event" | "slot",
    name: string,
    oldValue: any,
    newValue: any
  ): void;
  disconnectedCallback(): void;
}

declare interface HfcProps {
  attrs?: Record<string, any>;
  events?: Record<string, (args: any) => void>;
  slots?: Record<string, (container: HTMLElement, args: any) => void>;
}

export default function toVueComponent(
  HFC: typeof HyperFunctionComponent,
  lowercase?: boolean
) {
  const propKeys = Object.keys(HFC.interface.attrs || {});
  const eventKeys = Object.keys(HFC.interface.events || {});
  const slotKeys = Object.keys(HFC.interface.slots || {});

  const propKeyMap: Record<string, string> = {};
  const eventKeyMap: Record<string, string> = {};
  const slotKeyMap: Record<string, string> = {};
  if (lowercase) {
    propKeys.forEach((k) => (propKeyMap[k.toLowerCase()] = k));
    eventKeys.forEach(
      (k) =>
        (eventKeyMap["on" + k[0].toUpperCase() + k.slice(1).toLowerCase()] = k)
    );
    slotKeys.forEach((k) => (slotKeyMap[k.toLowerCase()] = k));
  } else {
    propKeys.forEach((k) => (propKeyMap[k] = k));
    eventKeys.forEach(
      (k) => (eventKeyMap["on" + k[0].toUpperCase() + k.slice(1)] = k)
    );
    slotKeys.forEach((k) => (slotKeyMap[k] = k));
  }

  const observePropKeys = Object.keys(propKeyMap);
  const observeEventKeys = Object.keys(eventKeyMap);
  const observeSlotKeys = Object.keys(slotKeyMap);

  return defineComponent({
    props: observePropKeys,
    setup(vueProps, { slots: vueSlots, attrs: vueAttrs }) {
      const container = ref<HTMLDivElement>();

      const events: Record<string, any> = {};
      observeEventKeys.forEach((key) => {
        const event = vueAttrs[key];
        if (!event) return;
        events[eventKeyMap[key]] = event;
      });

      const slots: Record<string, any> = {};

      interface RenderSlotItem {
        name: string;
        container: HTMLElement;
        slot: Function;
        args: any;
      }
      const renderedSlots: Record<string, RenderSlotItem> = reactive({});

      observeSlotKeys.forEach((key) => {
        const slot = vueSlots[key];
        if (!slot) return;
        slots[slotKeyMap[key]] = function (container: HTMLElement, args: any) {
          const slotItem = renderedSlots[key];
          if (!slotItem) {
            renderedSlots[key] = { name: key, container, slot, args };
            return;
          }

          if (!container) {
            delete renderedSlots[key];
            return;
          }

          slotItem.container = container;
          slotItem.args = args;
        };
      });

      let hfc: HyperFunctionComponent;
      onMounted(() => {
        const attrs: Record<string, any> = {};
        observePropKeys.forEach((key) => {
          if (vueProps[key]) attrs[propKeyMap[key]] = vueProps[key];
        });

        hfc = new HFC(container.value as HTMLDivElement, {
          attrs,
          events,
          slots,
        });
        hfc.connectedCallback();

        observePropKeys.forEach((key) => {
          watch(
            toRef(vueProps, key),
            (curr, prev) => {
              hfc.attributeChangedCallback("attr", propKeyMap[key], prev, curr);
            },
            { deep: true }
          );
        });
      });

      onUnmounted(() => {
        hfc.disconnectedCallback();
      });

      return () => [
        h("div", { ref: container }),
        Object.keys(renderedSlots).map((key) => {
          const item = renderedSlots[key];
          return h(Teleport, { to: item.container }, item.slot(item.args));
        }),
      ];
    },
  });
}
